#include "config.h"
#include <dirent.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "configio.h"
#include "defines.h"
#include "errcodes.h"
#include "fdbuf.h"
#include "itoa.h"
#include "list.h"

typedef list<mystring> slist;

#define fail(MSG) do { ferr.writeln(MSG); return false; } while(0)
#define fail_sys(MSG) do{ ferr.write(MSG); ferr.writeln(strerror(errno)); return false; }while(0)

struct remote
{
  mystring host;
  mystring proto;
  remote(const mystring& h, const mystring& p);
  ~remote();
};
remote::remote(const mystring& h, const mystring& p) : host(h), proto(p) { }
remote::~remote() { }
typedef list<remote> rlist;

static rlist remotes;
static int pausetime = 60;
static bool reload_config = false;

void catch_hup(int)
{
  signal(SIGHUP, catch_hup);
  reload_config = true;
}

bool load_remotes()
{
  slist rtmp;
  if(!config_readlist("remotes", rtmp) ||
     rtmp.count() == 0)
    return false;
  remotes.empty();
  for(slist::iter r(rtmp); r; r++) {
    int i = (*r).find(' ');
    if(i < 0)
      remotes.append(remote((*r), "smtp"));
    else
      remotes.append(remote((*r).left(i), (*r).right(i+1)));
  }
  return remotes.count() > 0;
}

bool load_config()
{
  bool result = true;
  if(!load_remotes()) result = false;
  
  if(!config_readint("pausetime", pausetime)) {
    pausetime = 60;
    result = false;
  }
  return result;
}

static slist files;
static bool reload_files = false;

void catch_alrm(int)
{
  signal(SIGALRM, catch_alrm);
  reload_files = true;
}

bool load_files()
{
  DIR* dir = opendir(".");
  if(!dir)
    fail_sys("Cannot open queue directory: ");
  files.empty();
  struct dirent* entry;
  while((entry = readdir(dir)) != 0) {
    const char* name = entry->d_name;
    if(name[0] == '.')
      continue;
    files.append(name);
  }
  closedir(dir);
  return true;
}

void exec_protocol(mystring filename, mystring hostname, mystring protocol)
{
  if(close(0) == -1 || close(1) == -1 || close(2) == -1)
    return;
  mystring program(PROTOCOL_DIR, protocol.c_str());
  execl(program.c_str(),
	program.c_str(), hostname.c_str(), filename.c_str(), 0);
}

#undef fail
#define fail(MSG) do { fout.writeln(MSG); return false; } while(0)
#define fail2(MSG1,MSG2) do{ fout.write(MSG1); fout.writeln(MSG2); return false; }while(0)

bool catchsender(pid_t pid)
{
  int status;
  if(waitpid(pid, &status, 0) == -1)
    fail("Error catching the child process return value.");
  else {
    if(WIFEXITED(status)) {
      status = WEXITSTATUS(status);
      if(status)
	fail2("Sending failed: ", errorstr[status]);
      else {
	fout.writeln("Sent file.");
	return true;
      }
    }
    else
      fail("Sending process crashed or was killed.");
  }
  return false;
}

bool send_one(mystring filename, mystring remote, mystring proto)
{
  fout.write("Starting delivery: protocol: ");
  fout.write(proto);
  fout.write(" host: ");
  fout.write(remote);
  fout.write(" file: ");
  fout.writeln(filename);
  pid_t pid = fork();
  switch(pid) {
  case -1:
    fail("Fork failed.");
  case 0:
    exec_protocol(filename, remote, proto);
    exit(ERR_EXEC_FAILED);
  default:
    if(!catchsender(pid))
      return false;
    if(unlink(filename.c_str()) == -1)
      fail("Can't unlink file.");
  }
  return true;
}

bool send_all()
{
  if(remotes.count() <= 0)
    fail("No remote hosts listed for delivery");
  if(files.count() == 0)
    return true;
  fout.write("Starting delivery, ");
  fout.write(itoa(files.count()));
  fout.writeln(" message(s) in queue.");
  for(rlist::iter remote(remotes); remote; remote++) {
    for(slist::iter file(files); file; files.remove(file)) {
      if(!send_one(*file, (*remote).host, (*remote).proto))
	break;
    }
  }
  fout.write("Delivery complete, ");
  fout.write(itoa(files.count()));
  fout.writeln(" message(s) remain.");
  return true;
}

static int trigger;

bool open_trigger()
{
  trigger = open(QUEUE_TRIGGER, O_RDONLY|O_NONBLOCK);
  if(trigger == -1)
    fail("Could not open trigger file.");
  return true;
}

bool read_trigger()
{
  if(trigger != -1) {
    char buf[1024];
    read(trigger, buf, sizeof buf);
    close(trigger);
  }
  return open_trigger();
}

bool do_select()
{
  fd_set readfds;
  FD_ZERO(&readfds);
  FD_SET(trigger, &readfds);
  struct timeval timeout;
  timeout.tv_sec = pausetime;
  timeout.tv_usec = 0;
  int s = select(trigger+1, &readfds, 0, 0,
		 (files.count() == 0) ? 0 : &timeout);
  if(s == 1) {
    fout.writeln("Trigger pulled.");
    read_trigger();
    reload_files = 1;
  }
  else if(s == -1 && errno != EINTR)
    fail("Internal error in select.");
  if(reload_config) {
    fout.writeln("Signal received, rescanning control files.");
    load_config();
  }
  if(reload_files) {
    fout.writeln("Rescanning queue.");
    load_files();
  }
  return true;
}

int main(int, char*[])
{
  if(!open_trigger())
    return 1;
  if(chdir(QUEUE_MSG_DIR) == -1) {
    fout.writeln("Could not chdir to queue message directory.");
    return 1;
  }
  
  signal(SIGALRM, catch_alrm);
  signal(SIGHUP, catch_hup);
  load_config();
  load_files();
  for(;;) {
    send_all();
    do_select();
  }
  return 0;
}
